import typing

__all__ = ["chunk", "getattributes"]

T = typing.TypeVar("T")


def chunk(iterable: typing.Iterable[T], size: int) -> typing.Iterator[typing.List[T]]:
    group = []
    for item in iterable:
        group.append(item)
        if len(group) == size:
            yield group
            group = []
    if group:
        message = (
            "Extra data in iterable; Iterable cannot split into "
            f"chunks of size {size} without extra data."
        )
        raise ValueError(message)


def getattributes(obj: object, *fields: str) -> object:
    for field in fields:
        obj = getattr(obj, field)
    return obj



from __future__ import annotations

import ctypes
import dataclasses
import enum
import functools
import typing

from hsc_instructions.sized_numbers import Uint8, Uint16, Uint24
from hsc_instructions.arg_types import (
    Label,
    Register,
    Shift,
    ShiftPointerDeref,
    Uint16PointerDeref,
)


def format_fields(
    **kwargs: int,
) -> typing.List[typing.Tuple[str, typing.Type[ctypes.c_uint], int]]:
    assert sum(kwargs.values()) == 24
    return [
        ("opcode", ctypes.c_uint, 8),
        *((name, ctypes.c_uint, value) for name, value in kwargs.items()),
    ]


class EncodingType(enum.Enum):
    # Don't have a good name for this encoding
    RRI = format_fields(rd=4, rm=4, imm16=16)

    # TODO: For shift, I am assming that the most significant bit determines
    # if it is a left shift (bit would be 0) or right shift (bit would be 1).
    # I don't know if that is actually true. That needs to be figured out.
    SHIFT = format_fields(rd=4, rm=4, rn=4, shift_type=1, shift=5, _padding=6)
    IMM24 = format_fields(imm24=24)
    REGISTER = format_fields(rd=4, rm=4, rn=4, _padding=12)
    JUMP = format_fields(rd=4, _padding=4, address=16)

    # Hacky solution, but it should work
    @functools.cached_property
    def encoding_class(self) -> typing.Type[BinaryInstruction]:
        class GenBinaryInstruction(BinaryInstruction):
            _fields_ = self.value

        return GenBinaryInstruction


class BinaryInstruction(ctypes.BigEndianStructure):
    opcode: int

    def to_bytes(self) -> bytes:
        return bytes(self)


@dataclasses.dataclass(init=False)
class EncodingSlot:
    arg_index: int
    # Allows multiple field accesses
    field: typing.Sequence[str]

    def __init__(self, arg_index: int, *fields: str) -> None:
        self.arg_index = arg_index
        self.field = fields

    def reversed_index(self) -> EncodingSlot:
        return type(self)(-self.arg_index - 1, *self.field)


@dataclasses.dataclass
class InstructionInfo:
    """
    Information about the instruction

    types: The corrosponding types of the instruction arguments
    encoding_slots: Dict of the slot on the encoding name and where to get its
    value in the instruction's args
    value: The argument's value in the encoding
    encoding: The instruction's encoding
    """

    types: typing.Sequence[typing.Type[InstructionArgTypes]]
    encoding_slots: typing.Dict[str, EncodingSlot]
    value: int
    encoding_type: EncodingType

    def __init__(
        self,
        types: typing.Sequence[typing.Type[InstructionArgTypes]],
        encoding_slots: typing.Dict[str, EncodingSlot],
        value: int,
        encoding_type: EncodingType,
    ) -> None:
        self.types = types
        self.encoding_slots = encoding_slots
        self.encoding_type = encoding_type
        self.value = Uint8(value)

    @classmethod
    def register_and_register_or_uint16(
        cls, instruction_value_register: int, instruction_value_imm: int
    ) -> MultInstructionInfo:
        return MultInstructionInfo(
            (
                cls.register_and_uint16(instruction_value_imm)[0],
                cls.two_register(instruction_value_register)[0],
            )
        )

    @classmethod
    def two_register_and_shift_or_uint16(
        cls, instruction_value_shift: int, instruction_value_imm: int
    ) -> MultInstructionInfo:
        return MultInstructionInfo(
            (
                cls.two_register_and_shift(instruction_value_shift)[0],
                cls(
                    (Register, Register, Uint16),
                    {
                        "rd": EncodingSlot(0),
                        "rm": EncodingSlot(1),
                        "imm16": EncodingSlot(2),
                    },
                    instruction_value_imm,
                    EncodingType.RRI,
                ),
            )
        )

    @classmethod
    def register_and_pointer_deref_with_imm_or_shift(
        cls, instruction_value_imm: int, instruction_value_shift: int
    ) -> MultInstructionInfo:
        return MultInstructionInfo(
            (
                cls(
                    (Register, Uint16PointerDeref),
                    {
                        "rd": EncodingSlot(0),
                        "rm": EncodingSlot(1, "register"),
                        "imm16": EncodingSlot(1, "increment"),
                    },
                    instruction_value_imm,
                    EncodingType.RRI,
                ),
                cls(
                    (Register, ShiftPointerDeref),
                    {
                        "rd": EncodingSlot(0),
                        "rm": EncodingSlot(1, "register"),
                        "rn": EncodingSlot(1, "increment", "register"),
                        "shift_type": EncodingSlot(1, "increment", "type"),
                        "shift": EncodingSlot(1, "increment", "amount"),
                    },
                    instruction_value_shift,
                    EncodingType.SHIFT,
                ),
            )
        )

    @classmethod
    def uint24(cls, instruction_value: int) -> MultInstructionInfo:
        return MultInstructionInfo(
            (
                cls(
                    (Uint24,),
                    {"imm24": EncodingSlot(0)},
                    instruction_value,
                    EncodingType.IMM24,
                ),
            )
        )

    @classmethod
    def two_register(cls, instruction_value: int) -> MultInstructionInfo:
        # TODO: I don't know the encoding for this, so I'm just
        # going to do REGISTER encoding
        return MultInstructionInfo(
            (
                cls(
                    (Register, Register),
                    {"rd": EncodingSlot(0), "rm": EncodingSlot(1)},
                    instruction_value,
                    EncodingType.REGISTER,
                ),
            )
        )

    @classmethod
    def two_register_and_shift(cls, instruction_value: int) -> MultInstructionInfo:
        return MultInstructionInfo(
            (
                cls(
                    (Register, Register, Shift),
                    {
                        "rd": EncodingSlot(0),
                        "rm": EncodingSlot(1),
                        "rn": EncodingSlot(2, "register"),
                        "shift_type": EncodingSlot(2, "type"),
                        "shift": EncodingSlot(2, "amount"),
                    },
                    instruction_value,
                    EncodingType.SHIFT,
                ),
            )
        )

    @classmethod
    def register_and_uint16(cls, instruction_value: int) -> MultInstructionInfo:
        return MultInstructionInfo(
            (
                cls(
                    (Register, Uint16),
                    {"rd": EncodingSlot(0), "imm16": EncodingSlot(1)},
                    instruction_value,
                    EncodingType.RRI,
                ),
            )
        )

    @classmethod
    def register_and_uint16_or_shift(
        cls, instruction_value_uint16: int, instrction_value_shift: int
    ) -> MultInstructionInfo:
        return MultInstructionInfo(
            (
                cls.register_and_uint16(instruction_value_uint16)[0],
                cls(
                    (Register, Shift),
                    {
                        "rd": EncodingSlot(0),
                        "shift_type": EncodingSlot(1, "type"),
                        "shift": EncodingSlot(1, "amount"),
                        "rn": EncodingSlot(1, "register"),
                    },
                    instrction_value_shift,
                    EncodingType.SHIFT,
                ),
            )
        )

    @classmethod
    def register_and_label(cls, instruction_value: int) -> MultInstructionInfo:
        return MultInstructionInfo(
            (
                cls(
                    (Register, Label),
                    {"rd": EncodingSlot(0), "address": EncodingSlot(1)},
                    instruction_value,
                    EncodingType.JUMP,
                ),
            )
        )

    @classmethod
    def label(cls, instruction_value: int) -> MultInstructionInfo:
        return MultInstructionInfo(
            (
                cls(
                    (Label,),
                    {"address": EncodingSlot(0)},
                    instruction_value,
                    EncodingType.JUMP,
                ),
            )
        )


class NoEncodingError(Exception):
    pass


class MultInstructionInfo(typing.Tuple[InstructionInfo]):
    types: typing.List[typing.FrozenSet[typing.Type[InstructionArgTypes]]]

    def __init__(self, args) -> None:
        # This will be an issue if different versions of the same
        # instruction have a different argument count
        self.types = [
            frozenset(instruction_type)
            for instruction_type in zip(*(instruction.types for instruction in self))
        ]

    def get_encoding(
        self, args: typing.Union[typing.Sequence[InstructionArgTypes], int]
    ) -> InstructionInfo:
        if isinstance(args, int):
            try:
                return next(op for op in self if op.value == args)
            except StopIteration:
                raise NoEncodingError(
                    f"No valid encoding for this instruction with op code {args}"
                ) from None
        for instruction in self:
            if len(instruction.types) == len(args) and all(
                isinstance(arg, arg_type)
                for arg, arg_type in zip(args, instruction.types)
            ):
                return instruction

        raise NoEncodingError("No valid encoding with those args with this instruction")

    def reverse(self) -> MultInstructionInfo:
        return MultInstructionInfo(
            dataclasses.replace(
                instruction,
                types=tuple(reversed(instruction.types)),
                encoding_slots={
                    key: value.reversed_index()
                    for key, value in instruction.encoding_slots.items()
                },
            )
            for instruction in self
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({super().__repr__()})"


class Instruction(enum.Enum):
    # The value arg in InstructionInfo is arbritrary right now; The actual value is TBD.
    LDR = InstructionInfo.register_and_pointer_deref_with_imm_or_shift(0, 1)
    STR = InstructionInfo.register_and_pointer_deref_with_imm_or_shift(2, 3).reverse()
    MOV = InstructionInfo.register_and_uint16_or_shift(4, 5)
    MOVT = InstructionInfo.register_and_uint16(6)
    ADD = InstructionInfo.two_register_and_shift_or_uint16(7, 8)
    ADDC = InstructionInfo.two_register_and_shift_or_uint16(9, 10)
    SUB = InstructionInfo.two_register_and_shift_or_uint16(11, 12)
    RSUB = InstructionInfo.two_register_and_shift_or_uint16(13, 14)
    SUBC = InstructionInfo.two_register_and_shift_or_uint16(15, 16)
    RSUBC = InstructionInfo.two_register_and_shift_or_uint16(17, 18)
    MUL = InstructionInfo.two_register_and_shift(19)
    AND = InstructionInfo.two_register_and_shift_or_uint16(20, 21)
    OR = InstructionInfo.two_register_and_shift_or_uint16(22, 23)
    XOR = InstructionInfo.two_register_and_shift_or_uint16(24, 25)
    NOT = InstructionInfo.two_register(26)  # Is this supposed to take two registers?
    CMP = InstructionInfo.register_and_register_or_uint16(27, 28)
    TST = InstructionInfo.register_and_register_or_uint16(29, 30)
    INT = InstructionInfo.uint24(31)
    JMP = InstructionInfo.label(32)
    # TODO: Add branch, jump, and other missed instructions

    @classmethod
    def from_int(cls, op: int) -> Instruction:
        try:
            return next(
                mem
                for mem in cls
                if op in {instruction.value for instruction in mem.value}
            )
        except StopIteration:
            raise ValueError(f"There is no instruction with the value {op}") from None


InstructionArgTypes = typing.Union[
    ShiftPointerDeref, Uint16PointerDeref, Uint16, Uint24, Shift, Label, Register
]



from __future__ import annotations

import sys
import typing

T = typing.TypeVar("T")


def format_error(error_message: str, line: typing.Optional[int]) -> str:
    line_msg = f"[line: {line}] " if line is not None else ""
    return f"{line_msg}{error_message}"


def error_print(*args, **kwargs) -> None:
    print(*args, file=sys.stderr, **kwargs)


class AsmException(Exception):
    """ Base class for exceptions in this module """

    def __init__(self, message: str, line: typing.Optional[int], *args):
        super().__init__(format_error(message, line), *args)

    @classmethod
    def collect_errors(cls, errors: typing.Iterable[str]) -> AsmException:
        """ Collects multiple error messages into a single error """
        message = "\n".join(errors)
        return cls(message, None)

    # Code is EX_DATAERR, taken from here https://man.openbsd.org/sysexits.3
    def error_exit(self, code=65) -> typing.NoReturn:
        """ Prints the error then exits the program """
        error_print(self.args[0])
        sys.exit(code)


class ScanError(AsmException):
    pass


class ParseError(AsmException):
    pass


class DecodingError(AsmException):
    pass


def exception_chain(
    functions: typing.Iterable[typing.Callable[..., T]],
    exception: BaseException,
    *args: object,
    **kwargs: object,
) -> T:
    for function in functions:
        try:
            return function(*args, **kwargs)
        except type(exception):
            pass
    raise exception


def merge_exceptions(
    name, *bases: typing.Type[BaseException]
) -> typing.Type[BaseException]:
    return type(name, bases, {})



from __future__ import annotations

import typing
from hsc_instructions.parsable import Parsable

MAX_NUM = 0b111111111111111111111111


class UintOverflowError(Exception):
    def __init__(self, number: int, max_size: int, type_name: str) -> None:
        message = (
            f"Number {number} is too large for type {type_name}. "
            f"Its max size is {max_size}."
        )
        super().__init__(message)


class UnderflowError(Exception):
    def __init__(self, number: int, type_name: str) -> None:
        super().__init__(
            f"Number {number} is too small (below zero) for type {type_name}."
        )


UintUnderflowError = UnderflowError


class PositiveSizedNumber(int):
    MAX: typing.ClassVar[int]

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        if self > self.MAX:
            raise UintOverflowError(self, self.MAX, cls.__name__)
        if self < 0:
            raise UnderflowError(self, cls.__name__)
        return self

    def __repr__(self) -> str:
        return f"{type(self).__name__}({super().__repr__()})"

    def __str__(self) -> str:
        return super().__repr__()

    @classmethod
    def from_int(cls, value: int) -> PositiveSizedNumber:
        return cls(value)

    @classmethod
    def from_fields(
        cls, fields: typing.Dict[typing.Tuple[str, ...], int]
    ) -> PositiveSizedNumber:
        return cls(fields[()])

    asm_code = __str__


class Uint5(PositiveSizedNumber):
    MAX = 0b11111


class Uint8(PositiveSizedNumber):
    MAX = 0b11111111


class Uint16(Parsable, PositiveSizedNumber):
    MAX = 0b1111111111111111


class Uint24(Parsable, PositiveSizedNumber):
    MAX = 0b111111111111111111111111



from __future__ import annotations

import dataclasses
import enum
import itertools
import typing

from hsc_instructions.errors import DecodingError
from hsc_instructions.instruction_info import Instruction as InstructionEnum
from hsc_instructions.arg_types import Label
from hsc_instructions.utilities import chunk, getattributes

if typing.TYPE_CHECKING:
    from instruction_info import InstructionArgTypes, InstructionInfo
    from sized_numbers import Uint16


__all__ = ["Instruction", "encode_instructions", "decode_instructions"]


@dataclasses.dataclass(init=False)
class Instruction:
    type: InstructionEnum
    # The exact types are based on the value of the InstructionEnum
    args: typing.Sequence[InstructionArgTypes]

    def __init__(
        self,
        instruction_type: InstructionEnum,
        args: typing.Sequence[InstructionArgTypes],
    ) -> None:
        # TODO: Have this raise an actual error, not just an assertion
        if __debug__:
            for arg, arg_type in zip(args, instruction_type.value.types):
                assert isinstance(arg, tuple(arg_type))
        self.type = instruction_type
        self.args = args

    def __str__(self) -> str:
        arg_list = ", ".join(arg.asm_code() for arg in self.args)
        return f"{self.type.name} {arg_list}"

    def to_bytes(self, symbol_table: typing.Dict[Label, Uint16]) -> bytes:
        encoding: InstructionInfo = self.type.value.get_encoding(self.args)
        encoding_values = {}
        for encoding_slot, field_name in encoding.encoding_slots.items():
            value = getattributes(self.args[field_name.arg_index], *field_name.field)
            if isinstance(value, enum.Enum):
                value = value.value
            elif isinstance(value, Label):
                value = symbol_table[value]
            encoding_values[encoding_slot] = value
        temp = encoding.encoding_type.encoding_class(encoding.value, **encoding_values)
        return temp.to_bytes()

    @classmethod
    def from_bytes(cls, byte_op: bytes) -> Instruction:
        if len(byte_op) != 4:
            raise DecodingError(
                f"Instructions are of length 4, this is of length {len(byte_op)}", None
            )
        op_code = byte_op[0]
        instruction: InstructionInfo = InstructionEnum.from_int(op_code)
        instruction_info = instruction.value.get_encoding(op_code)
        encoding_class = instruction_info.encoding_type.encoding_class
        byte_instruction = encoding_class.from_buffer_copy(byte_op)
        fields_to_values = (
            {
                tuple(decoded_rep.field): getattr(byte_instruction, encoded_field)
                for encoded_field, decoded_rep in encoding_slots
            }
            for _, encoding_slots in itertools.groupby(
                sorted(
                    instruction_info.encoding_slots.items(),
                    key=lambda x: x[1].arg_index,
                ),
                lambda x: x[1].arg_index,
            )
        )
        decoded_values = [
            arg_type.from_fields(fields)
            for arg_type, fields in zip(instruction_info.types, fields_to_values)
        ]
        return cls(instruction, decoded_values)


def encode_instructions(
    instructions: typing.Iterable[Instruction], symbol_table: typing.Dict[Label, Uint16]
) -> bytes:
    return b"".join(instruction.to_bytes(symbol_table) for instruction in instructions)


def decode_instructions(instructions: bytes) -> typing.Iterator[Instruction]:
    errors = []
    for i, instruction in enumerate(chunk(instructions, 4)):
        try:
            decoded_instruction = Instruction.from_bytes(bytes(instruction))
        except DecodingError as exc:
            errors.append(f"[instruction {i}] {exc.args[0]})")
        else:
            if not errors:
                yield decoded_instruction
    if errors:
        raise DecodingError.collect_errors(errors)



"""Enums for Tokens."""
from __future__ import annotations

import dataclasses
import enum
import typing

from hsc_instructions.errors import DecodingError, merge_exceptions
from hsc_instructions.sized_numbers import Uint5, Uint16
from hsc_instructions.parsable import Parsable, ParseMethod

__all__ = [
    "Register",
    "Shift",
    "ShiftType",
    "PointerDeref",
    "Label",
    "DecodeEnumFromValueMixin",
    "ShiftPointerDeref",
    "Uint16PointerDeref",
]


TC = typing.TypeVar("TC", bound="DecodeEnumFromValueMixin")


class DecodeEnumFromValueMixin(Parsable):
    @classmethod
    def from_int(cls: typing.Type[TC], value: int) -> TC:
        try:
            return next(
                enum_mem for enum_mem in cls if enum_mem.value == value  # type: ignore
            )
        except StopIteration:
            raise merge_exceptions("DecodeKeyError", KeyError, DecodingError)(
                f"No {cls.__name__} with encoding {value}"
            ) from None


class Register(DecodeEnumFromValueMixin, enum.IntEnum):
    """
    Enum of the registers.
    If the value is positive, it is the value of the register in code.
    """

    r0 = 0
    r1 = 1
    r2 = 2
    r3 = 3
    r4 = 4
    r5 = 5
    r6 = 6
    r7 = 7
    r8 = 8
    r9 = 9
    r10 = 10
    r11 = 11
    r12 = 12
    r13 = 13
    r14 = 14
    r15 = 15
    lr = 14
    sp = 15

    @classmethod
    def from_fields(cls, fields: typing.Dict[typing.Tuple[str, ...], int]) -> Register:
        return cls.from_int(fields[()])

    def asm_code(self) -> str:
        return self.name


# The values are the values left and right shifts are represented with
# in the encoding, which I do not know.
ShiftType = enum.IntEnum(  # type: ignore
    "ShiftType", {"<<": 0, ">>": 1}, type=DecodeEnumFromValueMixin
)


class Label(Parsable, str):
    @classmethod
    def from_fields(cls, fields: typing.Dict[typing.Tuple[str, ...], int]) -> Label:
        # Should the label's string representation be in hex?
        return cls(hex(fields[()]))

    def __repr__(self) -> str:
        return f"{type(self).__name__}({super().__repr__()})"

    def __str__(self) -> str:
        return super().__repr__()

    asm_code = __str__


@dataclasses.dataclass
class Shift(Parsable):
    type: ShiftType
    register: Register
    amount: Uint5

    @classmethod
    def from_fields(cls, fields: typing.Dict[typing.Tuple[str, ...], int]) -> Shift:
        shift_type = ShiftType.from_int(fields[("type",)])  # type: ignore
        shift_register = Register.from_int(fields[("register",)])
        shift_amount = Uint5(fields[("amount",)])
        return cls(shift_type, shift_register, shift_amount)

    def asm_code(self) -> str:
        return f"{self.register.asm_code()} {self.type.name} {self.amount.asm_code()}"


TI = typing.TypeVar("TI", Uint16, Shift)


@dataclasses.dataclass
class PointerDeref(Parsable, typing.Generic[TI]):
    register: Register
    increment: TI
    increment_parse: typing.ClassVar[ParseMethod]

    def asm_code(self) -> str:
        return f"[{self.register.asm_code()} + {self.increment.asm_code()}]"


# These are seperate because they have different binary encodings
class Uint16PointerDeref(PointerDeref[Uint16]):
    increment_parse = Uint16.parse

    @classmethod
    def from_fields(
        cls, fields: typing.Dict[typing.Tuple[str, ...], int]
    ) -> Uint16PointerDeref:
        dest_register = Register.from_int(fields[("register",)])
        increment = Uint16(fields[("increment",)])
        return cls(dest_register, increment)


class ShiftPointerDeref(PointerDeref[Shift]):
    increment_parse = Shift.parse

    @classmethod
    def from_fields(
        cls, fields: typing.Dict[typing.Tuple[str, ...], int]
    ) -> ShiftPointerDeref:
        dest_register = Register.from_int(fields[("register",)])
        shift = Shift.from_fields(
            {
                name[1:]: value
                for name, value in fields.items()
                if name[0] == "increment"
            }
        )
        return cls(dest_register, shift)



from __future__ import annotations

import typing

T = typing.TypeVar("T")
PT = typing.TypeVar("PT", bound="Parsable")

# cls: typing.Type[PT]
# tokens: typing.Union[typing.Sequence[T_contra], T_contra]
# line: int
# returns: PT
ParseFunction = typing.Callable[
    [typing.Type[PT], typing.Union[typing.Sequence[T], T], int], PT
]
ParseMethod = typing.Callable[[typing.Union[typing.Sequence[T], T], int], PT]

TPF = typing.TypeVar("TPF", bound=ParseFunction)


class Parsable:
    parse_function: typing.ClassVar[typing.Optional[ParseFunction]] = None

    @classmethod
    def make_parse_function(cls, function: TPF) -> TPF:
        cls.parse_function = function
        return function

    @classmethod
    def parse(
        cls: typing.Type[PT],
        tokens: typing.Union[typing.Sequence[typing.Any], typing.Any],
        line: int,
    ) -> PT:
        if cls.parse_function is None:
            decorator = f"{cls.__name__}.make_parse_function"
            msg = (
                "parse method not implemented; "
                f"specify parse function using {decorator} decorator"
            )
            raise NotImplementedError(msg)
        if not isinstance(tokens, typing.Sequence):
            tokens = [tokens]
        return cls.parse_function(cls, tokens, line)
